untyped

global function MpWeaponRocketLauncher_Init

global function OnWeaponActivate_weapon_rocket_launcher
global function OnWeaponDeactivate_weapon_rocket_launcher
global function OnWeaponPrimaryAttack_weapon_rocket_launcher
global function OnWeaponOwnerChanged_weapon_rocket_launcher


#if SERVER
global function OnWeaponNpcPrimaryAttack_weapon_rocket_launcher
global function OnWeaponNpcPrimaryAttack_S2S_weapon_rocket_launcher
global function rocketlaunch
global function CreateReplacementNpcForPlayer
#endif // #if SERVER


table<entity, float> replacementpilotnpchealth //the health for each replacementnpc (freeze breaks dying , so this is a janky workaround)
table<entity, entity> replacementpilotnpccreator //the creator for each replacementnpc, used to sync deaths
table<entity, entity> creatorreplacementpilotnpc //the replacementnpc for each creator , used to find


//14 //RUMBLE_FLAT_BOTH
const LOCKON_RUMBLE_INDEX 	= 1 //RUMBLE_PISTOL
const LOCKON_RUMBLE_AMOUNT	= 45
const S2S_MISSILE_SPEED = 2500
const S2S_MISSILE_HOMING = 5000

function MpWeaponRocketLauncher_Init()
{
	RegisterSignal( "StopLockonRumble" )
	RegisterSignal( "StopGuidedLaser" )

	PrecacheModel($"models/weapons/bullets/projectile_rocket_large.mdl")
}

function MissileThink( weapon, missile )
{
	expect entity( missile )

	#if SERVER
		missile.EndSignal( "OnDestroy" )

		bool playedWarning = false

		while ( IsValid( missile ) )
		{
			entity target = missile.GetMissileTarget()

			if ( IsValid( target ) && target.IsPlayer() )
			{
				float distance = Distance( missile.GetOrigin(), target.GetOrigin() )

				if ( distance < 1536 && !playedWarning )
				{
					EmitSoundOnEntityOnlyToPlayer( target, target, "titan_cockpit_missile_close_warning" )
					playedWarning = true
				}
			}

			WaitFrame()
		}
	#endif
}

void function OnWeaponActivate_weapon_rocket_launcher( entity weapon )
{
	if ( !( "initialized" in weapon.s ) )
	{
		weapon.s.missileThinkThread <- MissileThink
		weapon.s.initialized <- true
	}

	bool hasGuidedMissiles = weapon.HasMod( "guided_missile" )

	if ( !hasGuidedMissiles )
	{
		SmartAmmo_SetAllowUnlockedFiring( weapon, !IsMultiplayer() )
		if ( IsSingleplayer() )
		{
			SmartAmmo_SetMissileSpeed( weapon, 1750 )
			SmartAmmo_SetMissileHomingSpeed( weapon, 70 )

			if ( weapon.HasMod( "burn_mod_rocket_launcher" ) )
				SmartAmmo_SetMissileSpeedLimit( weapon, 1250 )
			else if ( weapon.HasMod( "sp_s2s_settings" ) )
			{
				SmartAmmo_SetMissileSpeedLimit( weapon, 9000 )
				SmartAmmo_SetMissileSpeed( weapon, S2S_MISSILE_SPEED )
				SmartAmmo_SetMissileHomingSpeed( weapon, S2S_MISSILE_HOMING )
			}
			else
				SmartAmmo_SetMissileSpeedLimit( weapon, 900 )
		}
		else
		{
			SmartAmmo_SetMissileSpeed( weapon, 1200 )
			SmartAmmo_SetMissileHomingSpeed( weapon, 125 )

			if ( weapon.HasMod( "burn_mod_rocket_launcher" ) )
				SmartAmmo_SetMissileSpeedLimit( weapon, 1300 )
			else
				SmartAmmo_SetMissileSpeedLimit( weapon, 1400 )
		}

		SmartAmmo_SetMissileShouldDropKick( weapon, false )  // TODO set to true to see drop kick behavior issues
		SmartAmmo_SetUnlockAfterBurst( weapon, true )
	}

	entity weaponOwner = weapon.GetWeaponOwner()

	if ( hasGuidedMissiles )
	{
		if ( !("guidedLaserPoint" in weaponOwner.s) )
			weaponOwner.s.guidedLaserPoint <- null

		thread CalculateGuidancePoint( weapon, weaponOwner )
	}
}

void function OnWeaponDeactivate_weapon_rocket_launcher( entity weapon )
{
	if ( weapon.HasMod( "guided_missile" ) )
		weapon.Signal( "StopGuidedLaser" )
}

////////////////////////////////////////////////////////Edited Things Start Here////////////////////////////////////////////////////////
var function OnWeaponPrimaryAttack_weapon_rocket_launcher( entity weapon, WeaponPrimaryAttackParams attackParams )
{
	#if SERVER
	entity player = weapon.GetOwner()
	if (player.RocketActive == false) {
		thread rocketlaunch(player)
		return 1 //plays shoot anim
	}
	return 0 //doesnt play shoot anim
	#endif
}

#if SERVER
void function rocketlaunch(entity player) {
	try {
		//rocket init
		player.RocketActive = true
		thread rocketduration(player)
		player.RocketLaunchPos = player.GetOrigin()

		vector lastrocketpos = <0,0,0>
		entity mover = CreateScriptMover (player.GetOrigin(),player.GetAngles())
		entity rocketmover = CreateScriptMover (player.GetOrigin(),player.GetAngles())

		player.DisableWeaponViewModel()
		player.ForceStand()
		player.SetTakeDamageType( DAMAGE_NO )
		player.kv.VisibilityFlags = ENTITY_VISIBLE_TO_NOBODY
		player.SetParent(mover)

		thread CreateReplacementNpcForPlayer(player,player.GetOrigin(),player.GetAngles())

		entity rocketprop = CreatePropDynamic($"models/weapons/bullets/projectile_rocket_large.mdl",player.EyePosition(),<0,player.GetAngles().y,0>,0)
		rocketprop.SetParent(rocketmover)

		TraceResults result

		//rocket control phase
		while (player.RocketActive == true) {
			WaitFrame()

			rocketmover.NonPhysicsMoveTo(player.EyePosition()+player.GetViewVector()*160/*40 infornt of the playermover*/+AnglesToUp(player.EyeAngles())* -70,0.1,0,0)
			rocketmover.NonPhysicsRotateTo(player.EyeAngles(),0.1,0,0)

			mover.NonPhysicsMoveTo(player.GetOrigin()+player.GetViewVector()*120,0.1,0,0)

			result = GetViewTrace(player)

			if(IsValid(result.hitEnt)) { //warning sound
				if(result.hitEnt.IsPlayer()) {
					if (result.hitEnt.RocketWarningActive == false) {
						EmitSoundOnEntityOnlyToPlayer(result.hitEnt,result.hitEnt,"titan_cockpit_missile_close_warning")
						result.hitEnt.RocketWarningActive = true
						thread SoundCooldown(result.hitEnt)
					}
				}
			}

			if (GetVectorDistance(result.endPos,player.EyePosition())<100) {
				break
			} else {
				lastrocketpos = mover.GetOrigin()
			}
		}

		//rocket explosion
		Explosion(lastrocketpos,player,null,100,2000,350,350,SF_ENVEXPLOSION_NOSOUND_FOR_ALLIES,lastrocketpos,100,damageTypes.explosive,eDamageSourceId.rocket,CLUSTER_ROCKET_FX_TABLE )
		EmitSoundOnEntityOnlyToPlayer(player,player,"explo_archer_impact_1p")
		if (IsValid(result.hitEnt) && result.hitEnt.IsPlayer()){ //cancel warning sound when rocket hits
			if (result.hitEnt.RocketWarningActive == true) {
				StopSoundOnEntity( result.hitEnt,"titan_cockpit_missile_close_warning" )
			}
		}


		//cleanup
		if (IsValid(creatorreplacementpilotnpc[player])) { //if there is a replacementnpc sync player to him , otherwise use rocketlaunchpos
			vector exitpos =  FindNearestSafeStandingPos(player,creatorreplacementpilotnpc[player].GetOrigin(),1)
			player.SetOrigin(exitpos)

		}
		else {player.SetOrigin(player.RocketLaunchPos)}

		player.ClearParent()
		player.UnforceStand()
		player.EnableWeaponViewModel()
		player.SetTakeDamageType( DAMAGE_YES )
		player.kv.VisibilityFlags = ENTITY_VISIBLE_TO_EVERYONE

		mover.Destroy()
		rocketprop.Destroy()

		player.RocketActive = false

	}catch(ex) {print(ex);print("damn")}
}


void function rocketduration(entity player) {

	for(int i = 0;i<50;i++) { //timer after wich the rocket explodes even without hitting a wall
		if (player.RocketActive == false) {return} //cancel if rocket is already inactive (due to impact) , this prevents timer overlap
		WaitFrame()
	}
	player.RocketActive = false
}

void function SoundCooldown(entity player) {
	wait 2
	if (IsValid(player)){player.RocketWarningActive = false}
}

void function CreateReplacementNpcForPlayer(entity player,vector location,vector angles=<0,0,0>) {
	WaitFrame();WaitFrame() //wait a bit for the rocket not to crash into the replacement


	int health = player.GetMaxHealth()
	entity replacementpilotnpc = CreateElitePilot( player.GetTeam(),location,angles)
	DispatchSpawn( replacementpilotnpc )

	replacementpilotnpc.SetModel( player.GetModelName() )
	replacementpilotnpc.SetTitle( player.GetPlayerName() )

	replacementpilotnpc.SetMaxHealth(health)
	replacementpilotnpc.SetHealth(health)

	replacementpilotnpc.ReplaceActiveWeapon(player.GetActiveWeapon().GetWeaponClassName())
	replacementpilotnpc.Anim_ScriptedPlay( "ACT_MP_STAND_IDLE" )

	replacementpilotnpchealth[replacementpilotnpc] <- float(health)
	replacementpilotnpccreator[replacementpilotnpc] <-player
	creatorreplacementpilotnpc[player] <- replacementpilotnpc

	AddEntityCallback_OnDamaged(replacementpilotnpc,replacementpilotnpcdamagethink)

	WaitFrame();WaitFrame() //give time for anim and weapon swap

	replacementpilotnpc.Freeze()

	thread replacementpilotnpcloopthink(replacementpilotnpc)
}

void function replacementpilotnpcdamagethink(entity ent,var damageInfo) {
	float damageAmount = DamageInfo_GetDamage( damageInfo )
	replacementpilotnpchealth[ent] = replacementpilotnpchealth[ent] - damageAmount

	if (replacementpilotnpchealth[ent] <= 0) { //npcreplacement destroyed
	   replacementpilotnpccreator[ent].SetTakeDamageType( DAMAGE_YES )
	   replacementpilotnpccreator[ent].TakeDamage(100,DamageInfo_GetAttacker( damageInfo ),null, {damageSourceId = DamageInfo_GetDamageSourceIdentifier( damageInfo )})
	   ent.Destroy()
	}
}

void function replacementpilotnpcloopthink(entity ent) {
	entity downmover = CreateScriptMover (ent.GetOrigin(),ent.GetAngles())
	ent.SetParent(downmover)
	while (IsValid(ent)) {
		WaitFrame()

		try{
			if (!PosInSolid(ent.GetOrigin()+ <0,0, -10>,ent)) {
				downmover.NonPhysicsMoveTo(ent.GetOrigin()+<0,0, -45>,0.1,0,0) //move down until on ground
			}
		}catch(ex) {print(ex)}

		if (!replacementpilotnpccreator[ent].RocketActive && IsValid(ent)) {
			ent.Destroy()
			break
		}

	}
}


///////////////////////////////////////////////////utility///////////////////////////////////////////////////////////////////
bool function PosInSolid (vector targetpos,entity ent = null) {
	try{
		vector mins
		vector maxs
		try {
			mins = ent.GetBoundingMins()
			maxs = ent.GetBoundingMaxs()
		} catch(ex0) {
			mins = < -20,-20,0>
			maxs = <20,20,1>
			print(ex0)
		}

		array ignoreent = []
		if (IsValid(ent)) {
			ignoreent.append(ent)
		}

		TraceResults result = TraceHull( targetpos ,targetpos, mins, maxs,ignoreent,TRACE_MASK_PLAYERSOLID, TRACE_COLLISION_GROUP_DEBRIS )
		if ( IsValid( result.hitEnt )  ){
			return true
		}

		return false
	}catch(ex){print(ex);return false}
}

float function GetVectorDistance (vector vector1 , vector vector2,) {
	float distance = pow((pow(vector1.x - vector2.x,2))+(pow(vector1.y - vector2.y,2))+(pow(vector1.z - vector2.z,2)),0.5)

	return distance
}

vector function FindNearestSafeStandingPos(entity player, vector targetPos, int severity) //this uses recursion I think
{
    if(severity >= 1500)
    {
        print("Error:could not find a save position for player")
       return < player.GetOrigin().x,player.GetOrigin().y,player.GetOrigin().z+1500 >
    }


	//check if adding/removing severity  from the x,y or z coords will stop the player from clipping

	//adding / removing z(height)
	if( !PosInSolid(< targetPos.x, targetPos.y, targetPos.z +severity> , player ) )
        return < targetPos.x, targetPos.y, targetPos.z +severity>

    if( !PosInSolid(< targetPos.x, targetPos.y, targetPos.z -severity > , player ) )
        return < targetPos.x, targetPos.y, targetPos.z -severity>

	//adding / removing y
    if( !PosInSolid(< targetPos.x, targetPos.y+severity, targetPos.z > , player) )
        return < targetPos.x, targetPos.y+severity, targetPos.z >

    if( !PosInSolid(< targetPos.x, targetPos.y-severity, targetPos.z > , player ) )
        return < targetPos.x, targetPos.y-severity, targetPos.z >

	//adding / removing x
    if( !PosInSolid(< targetPos.x+severity, targetPos.y, targetPos.z > , player ) )
        return < targetPos.x+severity, targetPos.y, targetPos.z >

    if( !PosInSolid(< targetPos.x-severity, targetPos.y, targetPos.z > , player ) )
        return < targetPos.x-severity, targetPos.y, targetPos.z >



    return FindNearestSafeStandingPos( player, targetPos, severity+5 ) //severity goes up by 5 each time adding severity doesnt stop the player from clipping
}




#endif









////////////////////////////////////////////////////////Edited Things End Here////////////////////////////////////////////////////////

#if SERVER
var function OnWeaponNpcPrimaryAttack_S2S_weapon_rocket_launcher( entity weapon, WeaponPrimaryAttackParams attackParams, entity target )
{
	entity weaponOwner = weapon.GetWeaponOwner()

	bool shouldPredict = false
	entity missile = weapon.FireWeaponMissile( attackParams.pos, attackParams.dir, S2S_MISSILE_SPEED, damageTypes.projectileImpact | DF_IMPACT, damageTypes.explosive, false, shouldPredict )

	if ( missile )
	{
		missile.kv.lifetime = 20
		missile.SetMissileTarget( target, < 0, 0, 0 > )
		missile.SetHomingSpeeds( S2S_MISSILE_HOMING, 0 )
	}
}

var function OnWeaponNpcPrimaryAttack_weapon_rocket_launcher( entity weapon, WeaponPrimaryAttackParams attackParams )
{
	/*
	// NPC can shoot the weapon at non-players, but when shooting at players it must be a titan
	entity owner = weapon.GetWeaponOwner()
	if ( IsValid( owner ) )
	{
		entity enemy = owner.GetEnemy()
		if ( IsValid( enemy ) )
		{
			if ( enemy.IsPlayer() && !enemy.IsTitan() )
				return
		}
	}
	*/

	weapon.EmitWeaponNpcSound( LOUD_WEAPON_AI_SOUND_RADIUS_MP, 0.2 )
	entity missile = weapon.FireWeaponMissile( attackParams.pos, attackParams.dir, 1800.0, damageTypes.projectileImpact, damageTypes.explosive, false, PROJECTILE_NOT_PREDICTED )
	if ( missile )
	{
		missile.InitMissileForRandomDriftFromWeaponSettings( attackParams.pos, attackParams.dir )
		if ( weapon.w.missileFiredCallback != null )
		{
			weapon.w.missileFiredCallback( missile, weapon.GetWeaponOwner() )
		}
	}
}
#endif // #if SERVER

//GUIDED MISSILE FUNCTIONS
function CalculateGuidancePoint( entity weapon, entity weaponOwner )
{

	weaponOwner.EndSignal( "OnDestroy" )
	weapon.EndSignal( "OnDestroy" )
	weapon.EndSignal( "StopGuidedLaser" )

	entity info_target
	#if SERVER
		info_target = CreateEntity( "info_target" )
		info_target.SetOrigin( weapon.GetOrigin() )
		info_target.SetInvulnerable()
		DispatchSpawn( info_target )
		weapon.s.guidedMissileTarget <- info_target
	#endif

	OnThreadEnd(
		function() : ( weapon, info_target )
		{
			if ( IsValid( info_target ) )
			{
				info_target.Kill_Deprecated_UseDestroyInstead()
				delete weapon.s.guidedMissileTarget
			}
		}
	)

	while ( true )
	{
		if ( !IsValid_ThisFrame( weaponOwner ) || !IsValid_ThisFrame( weapon ) )
			return

		weaponOwner.s.guidedLaserPoint = null
		if ( weapon.IsWeaponInAds())
		{
			TraceResults result = GetViewTrace( weaponOwner )
			weaponOwner.s.guidedLaserPoint = result.endPos
			#if SERVER
				info_target.SetOrigin( result.endPos )
			#endif
		}

		WaitFrame()
	}

}

function InitializeGuidedMissile( entity weaponOwner, entity missile )
{
		missile.s.guidedMissile <- true
		if ( "missileInFlight" in weaponOwner.s )
			weaponOwner.s.missileInFlight = true
		else
			weaponOwner.s.missileInFlight <- true

		missile.kv.lifetime = 10

		#if SERVER
			missile.SetOwner( weaponOwner )
			thread playerHasMissileInFlight( weaponOwner, missile )
		#endif
}

#if SERVER
function playerHasMissileInFlight( entity weaponOwner, entity missile )
{
	weaponOwner.EndSignal( "OnDestroy" )

	OnThreadEnd(
		function() : ( weaponOwner )
		{
			if ( IsValid( weaponOwner ) )
			{
				weaponOwner.s.missileInFlight = false
				//Using a remote call because if this thread is on the client it gets triggered prematurely due to prediction.
				Remote_CallFunction_NonReplay( weaponOwner, "ServerCallback_GuidedMissileDestroyed" )
			}
		}
	)

	WaitSignal( missile, "OnDestroy" )
}
#endif // SERVER


void function OnWeaponOwnerChanged_weapon_rocket_launcher( entity weapon, WeaponOwnerChangedParams changeParams )
{
	#if SERVER
		weapon.w.missileFiredCallback = null
	#endif
}